from Crypto.Cipher import AES
from typing import List, Union, Any, Optional, Dict
from pathlib import PurePath
from .bitcoinj_compat import *
import hashlib
import arrow
from coincurve import PrivateKey, PublicKey
import base58
import ipaddress
import json
import base64
from .crypto import ripemd160_sha256
from .exceptions import SerializationError, PasswordError
from hashlib import md5

class ProtobufWallet(object):

    def __init__(self, data: Any) -> None:
        self.raw_data = data
        self.wallet = Wallet()
        self.default_wifnetwork = 0
        self.keypairs = []
        self.description = ""
        self.pool = []
        self.txes = []
        self.mnemonics = []
        self.bestblock = {}
        self.tags = []
        self.extensions = []
        self.lastblockhash = None
        self.encryption_params = None
        self.encrypted = False

    @classmethod
    def load(cls, filename: Union[str, PurePath]):
        with open(filename, "rb") as d:
            data = d.read()
        if b'org.' in data:
            return cls(data)
        else:
            data = base64.b64decode(data)


        return cls(data)

    def parse(self, passphrase: Optional[str]=None):
        if self.raw_data[:8] == b'Salted__':
            if passphrase is None:
                raise PasswordError
            else:
                salt = self.raw_data[8:16]
                newdata = b''
                tmp2 = b''
                tmp = passphrase.encode() + salt
                while len(newdata) < 32 + AES.block_size:
                    msg = tmp2 + tmp
                    tmp2 = md5(msg).digest()
                    newdata += tmp2
                key = newdata[:32]
                iv = newdata[32:32 + AES.block_size]
                cipher = AES.new(key, AES.MODE_CBC, iv)
                try:
                    padded_plain = cipher.decrypt(self.raw_data[AES.block_size:])
                    pad_len = padded_plain[-1]
                    if isinstance(pad_len, str):
                        pad_len = ord(pad_len)
                    plain_data = padded_plain[:-pad_len]
                    self.wallet.parse(plain_data)
                except:
                    raise PasswordError
        else:
            self.wallet.parse(self.raw_data)

        if self.wallet.network_identifier == "org.dogecoin.production":
            self.default_wifnetwork = 30
        elif self.wallet.network_identifier == "org.bitcoin.production":
            self.default_wifnetwork = 0
        elif self.wallet.network_identifier == "org.litecoin.production":
            self.default_wifnetwork = 48
        elif self.wallet.network_identifier == "org.feathercoin.production":
            self.default_wifnetwork = 14
        else:
            self.default_wifnetwork = 0
            """
            Not sure what other variants have been made into bitcoinj-compatible
            """
        self.encryption_params = self.wallet.encryption_parameters.to_dict()
        if self.wallet.encryption_type > 1:
            self.encrypted = True
        if passphrase is None:
            for k in self.wallet.key:
                if k.type == 3:
                    mnemonic = k.secret_bytes.decode()
                    deterministic_seed = k.deterministic_seed.hex()
                    self.mnemonics.append({"mnemonic": mnemonic, "deterministic_seed": deterministic_seed})
                elif k.type == 4 or k.type == 1:
                    creation_timestamp = arrow.get(int(k.creation_timestamp)).isoformat()
                    keydict = {"creation_timestamp": creation_timestamp}
                    if k.type == 4:
                        deterministic_key = k.deterministic_key.to_dict()
                        deterministic_key["chainCode"] = base64.b64decode(deterministic_key['chainCode']).hex()
                        keydict["deterministic_key"] = deterministic_key
                    prefix = bytes([self.default_wifnetwork])
                    pubkey = base58.b58encode_check(prefix + ripemd160_sha256(k.public_key)).decode()
                    keydict["pubkey"] = pubkey
                    if len(k.secret_bytes) > 0:
                        privkey_prefix = int.to_bytes(self.default_wifnetwork + 128, 1, byteorder='big')
                        compressed = base58.b58encode_check(privkey_prefix + k.secret_bytes).decode()
                        uncompressed = base58.b58encode_check(privkey_prefix + k.secret_bytes + b'\x01').decode()
                        keydict["compressed_private_key"] = compressed
                        keydict["uncompressed_private_key"] = uncompressed
                    keydict["label"] = k.label
                    self.keypairs.append(keydict)

            if hasattr(self.wallet, 'lastSeenBlockHash'):
                self.lastblockhash = self.wallet.last_seen_block_hash.hex()
            if hasattr(self.wallet, 'transaction'):
                for t in self.wallet.transaction:
                    tx = t.to_dict()
                    tx["hash"] = base64.b64decode(tx['hash']).hex()
                    tx["updatedAt"] = arrow.get(int(tx["updatedAt"])).isoformat()
                    txinputs = []
                    for txinput in tx['transactionInput']:
                        inputdict = {}
                        if 'transactionOutPointHash' in txinput.keys():
                            inputdict['transactionOutPointHash'] = base64.b64decode(txinput['transactionOutPointHash']).hex()
                        if 'transactionOutPointIndex' in txinput.keys():
                            inputdict['transactionOutPointIndex'] = txinput['transactionOutPointIndex']
                        if 'scriptBytes' in txinput.keys():
                            inputdict['scriptBytes'] = base64.b64decode(txinput['scriptBytes']).hex()
                        txinputs.append(inputdict)
                    tx['transactionInput'] = txinputs
                    txoutputs = []
                    for txoutput in tx['transactionOutput']:
                        outputdict = {}
                        if 'spentByTransactionHash' in txoutput.keys():
                            outputdict['spentByTransactionHash'] = base64.b64decode(
                                    txoutput['spentByTransactionHash']).hex()
                        if 'spentByTransactionIndex' in txoutput.keys():
                            outputdict['spentByTransactionIndex'] = txoutput['spentByTransactionIndex']
                        if 'scriptBytes' in txoutput.keys():
                            outputdict['scriptBytes'] = base64.b64decode(txoutput['scriptBytes']).hex()
                        if 'value' in txoutput.keys():
                            outputdict['value'] = txoutput['value']
                        txoutputs.append(outputdict)
                    tx['transactionOutput'] = txoutputs
                    bhashes = []
                    for bhash in tx['blockHash']:
                        bhashes.append(base64.b64decode(bhash).hex())
                    tx['blockHash'] = bhashes
                    self.txes.append(tx)
            self.description = self.wallet.description
            self.extensions = self.wallet.extension
            self.tags = self.wallet.tags
        else:
            self.keypairs = self.wallet.to_dict()['key']

    def dump_all(self, filepath: Optional[str] = None) -> Dict:
        output_items = {"keys": self.keypairs, "tx": self.txes, "mnemonics": self.mnemonics, "description": self.description}
        if filepath is not None:
            with open(filepath, "a") as ft:
                ft.write(json.dumps(output_items, sort_keys=True, indent=4))
        return output_items

    def dump_keys(self, filepath: Optional[str] = None) -> List:
        output_keys = []
        for each in self.mnemonics:
            output_keys.append(each)
        for each in self.keypairs:
            output_keys.append(each)
        if filepath is not None:
            with open(filepath, "a") as ft:
                ft.write(json.dumps(output_keys, sort_keys=True, indent=4))
        return output_keys












import collections
import datetime
import json
import socket
from typing import Dict, List, Optional

from bsddb3.db import *
from coincurve import PrivateKey, PublicKey
import base58
from .crypto import Crypter, doublesha256, doublesha256_checksum, ripemd160_sha256
from .exceptions import *
from .utils import BCDataStream, parse_CAddress, parse_BlockLocator, privkey_to_secret
import ipaddress

class Walletdat(object):
    def __init__(self, db: collections.OrderedDict) -> None:
        self.db_parsed = db
        self.keypairs = []
        self.rawkeys = []
        self.pool = []
        self.txes = []
        self.keymetas = []
        self.defaultkey = None
        self.default_wifnetwork = None
        self.addressbook = []
        self.version = 0
        self.minversion = 0
        self.setting = {}
        self.hdchain = {}
        self.bestblock = {}
        self.destdata = []
        self.orderposnext = 0
        self.flags = 0
        self.decrypter = Crypter()
        self.mkey = None
        self.encrypted = False

    @classmethod
    def load(cls, filename):
        """

        :param filename: name of the file to load, file should be a BerkeleyDB Wallet.dat file
        :type filename: string
        :return: cls
        :rtype:Walletdat
        """
        try:
            db = DB()
            db.open(filename, "main", DB_BTREE, DB_THREAD | DB_RDONLY)
            coll = collections.OrderedDict((k, db[k]) for k in db.keys())
            return cls(coll)
        except (DBNoSuchFileError, DBError):
            raise DatabaseError(file=filename)

    def parse(self, passphrase: Optional[str] = None) -> None:
        """Parse the raw bytes of the db's contents. A bit of a mess right now so API likely to change

        :param passphrase: Passphrase to the wallet
        :type passphrase: string
        :return: None 
        :rtype: None
        """
        for key, value in self.db_parsed.items():
            kds = BCDataStream(key)
            vds = BCDataStream(value)
            keytype = kds.read_string().decode()
            if keytype == "key":
                try:
                    keypair = KeyPair.parse_fromwallet(kds, vds)
                    self.keypairs.append(keypair)
                except KeypairError:
                    print(
                        "Error: Pubkey data doesn't match pubkey derived from private key"
                    )
            elif keytype == "wkey":
                keypair = KeyPair.parse_fromwallet(kds, vds)
                created = vds.read_int64()
                expires = vds.read_int64()
                comment = vds.read_string().decode()
                keypair.parse_wkeyinfo(created, expires, comment)
                self.keypairs.append(keypair)
            elif keytype == "keymeta":
                pubkey = kds.read_bytes(kds.read_compact_size())
                if len(pubkey) == 33:
                    compressed = True
                else:
                    compressed = False
                version = vds.read_int32()
                createtime = vds.read_int64()
                if version != 10:
                    hdkeypath = "No HD Key Found"
                    hdmasterkey = None
                else:
                    hdkeypath = vds.read_string().decode("utf-8")
                    hdmasterkey = vds.read_bytes(20).hex()
                if version != 12:
                    fingerprint = None
                    has_keyorigin = False
                else:
                    fingerprint = vds.read_uint32()
                    has_keyorigin = vds.read_boolean()
                for key in self.keypairs:
                    if key.publickey == PublicKey(pubkey).format(compressed=compressed):
                        key.set_keymeta(
                            version,
                            createtime,
                            hdkeypath,
                            hdmasterkey,
                            fingerprint,
                            has_keyorigin,
                        )
                else:
                    self.keymetas.append(
                        {
                            "version": version,
                            "createtime": createtime,
                            "hdkeypath": hdkeypath,
                            "hdmasterkey": hdmasterkey,
                            "fingerprint": fingerprint,
                            "has_keyorigin": has_keyorigin,
                        }
                    )
                break
            elif keytype == "defaultkey":
                pk = vds.read_bytes(vds.read_compact_size())
                if len(pk) == 33:
                    self.defaultkey = PublicKey(pk).format()
                else:
                    self.defaultkey = PublicKey(pk).format(compressed=False)
            elif keytype == "name":
                if len(self.addressbook) > 0:
                    addr = kds.read_string().decode("utf-8")
                    if any(item["address"] == addr for item in self.addressbook):
                        for item in self.addressbook:
                            if item["address"] == addr:
                                item.update(
                                    {"label": vds.read_string().decode("utf-8")}
                                )
                    else:
                        self.addressbook.append(
                            {
                                "address": addr,
                                "label": vds.read_string().decode("utf-8"),
                            }
                        )
                else:
                    addr = kds.read_string().decode("utf-8")
                    self.addressbook.append(
                        {"address": addr, "label": vds.read_string().decode("utf-8")}
                    )
                self.default_wifnetwork = ord(base58.b58decode_check(addr)[:1])
            elif type == "purpose":
                if len(self.addressbook) > 0:
                    addr = kds.read_string().decode("utf-8")
                    if any(item["address"] == addr for item in self.addressbook):
                        for item in self.addressbook:
                            if item["address"] == addr:
                                item.update(
                                    {"purpose": vds.read_string().decode("utf-8")}
                                )
                    else:
                        self.addressbook.append(
                            {
                                "address": addr,
                                "purpose": vds.read_string().decode("utf-8"),
                            }
                        )
                else:
                    addr = kds.read_string().decode("utf-8")
                    self.addressbook.append(
                        {"address": addr, "purpose": vds.read_string().decode("utf-8")}
                    )

            elif keytype == "tx":
                # todo: add segwit
                try:
                    txid = invert_txid(kds.read_bytes(32))
                    self.txes.append(Transaction.parse(txid, vds))
                except:
                    pass
            elif keytype == "hdchain":
                version = vds.read_uint32()
                chain_counter = vds.read_uint32()
                master_keyid = vds.read_bytes(20).hex()
                self.hdchain = {
                    "version": version,
                    "chain_counter": chain_counter,
                    "master_keyid": master_keyid,
                }
                if version > 2:
                    self.hdchain["internal_counter"] = vds.read_uint32()
            elif keytype == "version":
                self.version = vds.read_uint32()
            elif keytype == "minversion":
                self.minversion = vds.read_uint32()
            elif keytype == "setting":
                setname = kds.read_string().decode()
                if setname[0] == "f":
                    value = vds.read_boolean()
                elif setname == "addrIncoming":
                    value = str(ipaddress.ip_address(vds.read_bytes(4)))
                elif setname.startswith("addr"):
                    d = {"ip": "0.0.0.0", "port": 0, "nTime": 0}
                    try:
                        d["nVersion"] = vds.read_int32()
                        d["nTime"] = vds.read_uint32()
                        d["nServices"] = vds.read_uint64()
                        d["pchReserved"] = vds.read_bytes(12)
                        d["ip"] = str(ipaddress.ip_address(vds.read_bytes(4)))
                        d["port"] = vds.read_uint16()
                    except Exception:
                        pass
                    value = d
                elif setname == "nTransactionFee":
                    value = vds.read_int64()
                elif setname == "nLimitProcessors":
                    value = vds.read_int32()
                self.setting[setname] = value
            elif keytype == "bestblock":
                version = vds.read_int32()
                hashes = []
                for _ in range(vds.read_compact_size()):
                    hashes.append(vds.read_bytes(32).hex())
                self.bestblock = {"version": version, "hashes": hashes}
            elif keytype == "bestblock_nomerkle":
                version = vds.read_int32()
                hashes = []
                self.bestblock = {"version": version, "hashes": hashes}
            elif keytype == "pool":
                n = kds.read_int64()
                nversion = vds.read_int32()
                ntime = vds.read_int64()
                pubkey = vds.read_bytes(vds.read_compact_size())
                if len(pubkey) == 33:
                    compressed = True
                else:
                    compressed = False
                self.pool.append(
                    {
                        "n": n,
                        "nversion": nversion,
                        "ntime": ntime,
                        "publickey": PublicKey(pubkey).format(compressed=compressed),
                    }
                )
            elif keytype == "destdata":
                publickey = kds.read_string().decode()
                key = kds.read_string().decode()
                # destination = vds.read_string().decode()
                self.destdata.append({"publickey": publickey, "key": key})
            elif keytype == "orderposnext":
                self.orderposnext = vds.read_int64()
            elif keytype == "flags":
                self.flags = vds.read_uint64()
            elif keytype == "mkey":
                nid = kds.read_uint32()
                encrypted_key = vds.read_string()
                salt = vds.read_string()
                derivationmethod = vds.read_uint32()
                derivationiters = vds.read_uint32()
                self.mkey = {
                    "nID": nid,
                    "encrypted_key": encrypted_key,
                    "salt": salt,
                    "derivationmethod": derivationmethod,
                    "derivationiterations": derivationiters,
                }
                self.encrypted = True

            elif keytype == "ckey":
                publickey = kds.read_bytes(kds.read_compact_size())
                encrypted_privkey = vds.read_bytes(vds.read_compact_size())
                self.rawkeys.append({"publickey": publickey, "encrypted_privkey": encrypted_privkey})
            else:
                print("{} type not implemented".format(type))
        if self.encrypted:
            if passphrase is not None:
                self.decrypter.keyfrompassphrase(bytes(passphrase, "utf-8"), self.mkey["salt"], self.mkey["derivationiterations"], self.mkey["derivationmethod"])
                masterkey = self.decrypter.decrypt(self.mkey["encrypted_key"])
                self.decrypter.SetKey(masterkey)
                for rawkey in self.rawkeys:
                    try:
                        self.decrypter.SetIV(doublesha256(rawkey["publickey"]))
                        dec = self.decrypter.decrypt(rawkey["encrypted_privkey"])
                        self.keypairs.append(
                            KeyPair.parse_fromckey(
                            pubkey=rawkey["publickey"],
                            privkey=dec,
                            encryptedkey=rawkey["encrypted_privkey"],
                            crypted=False,
                        ))
                    except:
                        raise PasswordError
            else:
                print("No passphrase set for encrypted wallet")
                for rawkey in self.rawkeys:
                    self.keypairs.append(
                        KeyPair.parse_fromckey(
                            pubkey=rawkey["publickey"],
                            privkey=None,
                            encryptedkey=rawkey["encrypted_privkey"],
                        )
                    )


    def dump_keys(
        self,
        filepath: Optional[str] = None,
        version: Optional[int] = None,
        privkey_prefix_override: Optional[int] = None,
        compression_override: Optional[bool] = None,
    ) -> List:
        """ Dump just pubkey:privatekey either as a list, write to a file, or both.

        
        :param filepath: The output file. Leave as None to not write to file
        :type filepath:  string
        :param version: Version byte for the p2pkh key being generated. Should be between 0 and 127
        :type version:  int
        :param privkey_prefix_override: WIF Version bytes for secret. Should be between 128 and 255
        :type privkey_prefix_override: int
        :return: List of dicts with pubkey and privkey in bytesm or if not decrypted, pubkey and encrypted key
        :rtype: List
        :param compression_override: Compression setting for output keys, use None to automatically determine
        :type compression_override: bool
        """
        if len(self.keypairs) == 0 and self.db_parsed is not None:
            self.parse()
            # Run parse to populate if forgot
        output_list = []
        if version is None:
            prefix = self.default_wifnetwork
        else:
            prefix = version

        for keypair in self.keypairs:
            if privkey_prefix_override is not None:
                wif_prefix = privkey_prefix_override - 128
            else:
                wif_prefix = prefix
            pkey = keypair.pubkey_towif(prefix)
            if compression_override is not None:
                priv = keypair.privkey_towif(
                    wif_prefix, compressed=compression_override
                )
            else:
                priv = keypair.privkey_towif(wif_prefix, compressed=keypair.compressed)
            output_list.append({"public_key": pkey, "private_key": priv})

            if filepath is not None:
                with open(filepath, "a") as fq:
                    if self.mkey is None:
                        fq.write(pkey.decode() + ":" + priv.decode() + "\n")
                    else:
                        fq.write(
                            pkey.decode() + ":" + keypair.encryptedkey.hex() + "\n"
                        )

        return output_list

    def dump_all(
        self,
        filepath: Optional[str] = None,
        version: Optional[int] = None,
        privkey_prefix_override: Optional[int] = None,
    ) -> Dict:
        """ Dump all data from wallet

        :param filepath: The output file. Leave as None to not write to file
        :type filepath:  String
        :param version: Version byte for the p2pkh key being generated. Should be between 0 and 127
        :type version: Int
        :param privkey_prefix_override:  WIF Version byte override value just for the private key
        :type privkey_prefix_override: Int
        :return: A dict with the following key:values - keys: lists of dicts with compressed_private_key,
        public_key, uncompressed_private_key, label, created; pool: list of dicts with keys: n, nversion, ntime,
        publickey; tx: list of dicts with keys: txid, txin, txout, locktime; minversion, bestblock,
        default_network_version, orderposnext
        :rtype:
        """
        if len(self.keypairs) == 0 and self.db_parsed is not None:
            self.parse()

        structures = {
            "keys": [],
            "pool": [],
            "tx": [],
            "minversion": self.minversion,
            "version": self.version,
            "bestblock": self.bestblock,
            "default_network_version": self.default_wifnetwork,
            "orderposnext": self.orderposnext,
        }

        if version is None:
            prefix = self.default_wifnetwork
        else:
            prefix = version

        for keypair in self.keypairs:
            pkey = keypair.pubkey_towif(prefix)
            if keypair.encryptedkey is None:
                if privkey_prefix_override is not None:
                    wif_prefix = privkey_prefix_override - 128
                else:
                    wif_prefix = prefix
                priv_compressed = keypair.privkey_towif(wif_prefix, compressed=True)
                priv_uncompressed = keypair.privkey_towif(wif_prefix, compressed=False)
                keyd = {
                    "public_key": pkey.decode(),
                    "compressed_private_key": priv_compressed.decode(),
                    "uncompressed_private_key": priv_uncompressed.decode(),
                }
            else:
                priv_encrypted = keypair.encryptedkey.hex()
                keyd = {
                    "public_key": pkey.decode(),
                    "encrypted_private_key": priv_encrypted,
                }
            if len(self.addressbook) > 0:
                for a in self.addressbook:
                    if a["address"] == pkey.decode():
                        keyd["label"] = a["label"]
                        if "purpose" in a.keys():
                            keyd["purpose"] = a["purpose"]
            if keypair.createtime > 0:
                keyd["created"] = datetime.datetime.utcfromtimestamp(
                    keypair.createtime
                ).isoformat()
            structures["keys"].append(keyd)

        for tx in self.txes:
            apg = {
                "txid": tx.txid,
                "txin": tx.txin,
                "txout": tx.txout,
                "locktime": tx.locktime,
            }
            structures["tx"].append(apg)

        z = bytes([prefix])
        pools = []
        for p in self.pool:
            pools.append(
                {
                    "n": p["n"],
                    "nversion": p["nversion"],
                    "ntime": datetime.datetime.utcfromtimestamp(p["ntime"]).isoformat(),
                    "public_key": base58.b58encode_check(
                        z + ripemd160_sha256(p["publickey"])
                    ).decode(),
                }
            )

        sorted(pools, key=lambda i: (i["n"], i["ntime"]))
        structures["pool"] = pools

        defkey = base58.b58encode_check(z + ripemd160_sha256(self.defaultkey)).decode()
        structures["default_key"] = defkey

        if filepath is not None:
            with open(filepath, "a") as fq:
                fq.write(json.dumps(structures, sort_keys=True, indent=4))
        return structures


class KeyPair(object):
    """Keypair object, should not be called directly


    """

    def __init__(
        self, rawkey, rawvalue, pubkey, sec, compressed, privkey=None, encryptedkey=None
    ):
        self.rawkey = rawkey
        self.rawvalue = rawvalue
        self.publickey = pubkey
        self.privkey = privkey
        self.secret = sec
        self.version = None
        self.createtime = 0
        self.hdkeypath = None
        self.hdmasterkey = None
        self.compressed = compressed
        self.fingerprint = None
        self.has_keyorigin = False
        self.encryptedkey = encryptedkey
        self.expiretime = 0
        self.comment = ""

    @classmethod
    def parse_fromwallet(cls, kds, vds):
        """Class method to parse entry from wallet entry

        :param kds: BCDatastream object for keys
        :type kds: BCDataStream
        :param vds: BCDataStream object for values
        :type vds: BCDataStream
        :return: KeyPair
        :rtype: KeyPair
        """
        pubkeyraw = kds.read_bytes(kds.read_compact_size())
        privkeyraw = vds.read_bytes(vds.read_compact_size())
        if len(privkeyraw) == 279:
            sec = privkeyraw[9:41]
        else:
            sec = privkeyraw[8:40]
        privkey = PrivateKey(sec)
        pubkey = PublicKey(pubkeyraw)
        if len(pubkeyraw) == 33:
            compress = True
        else:
            compress = False
        if pubkey == privkey.public_key:
            pubkey = privkey.public_key.format(compressed=compress)
            return cls(
                rawkey=pubkeyraw,
                rawvalue=privkeyraw,
                pubkey=pubkey,
                privkey=privkey,
                sec=sec,
                compressed=compress,
            )
        else:
            raise KeypairError(message="Pubkey {} error".format(pubkey.format(compressed=compress).hex()))

    def parse_wkeyinfo(self, createtime: int, expiretime: int, comment: str) -> None:
        self.createtime = createtime
        self.expiretime = expiretime
        self.comment = comment

    @classmethod
    def parse_fromckey(cls, pubkey, privkey, encryptedkey, crypted=True):
        """Parse keypair from ckey (encrypted) values from wallet

        :param pubkey:
        :type pubkey:
        :param privkey:
        :type privkey:
        :param encryptedkey:
        :type encryptedkey:
        :param crypted:
        :type crypted:
        :return:
        :rtype:
        """
        pkey = PublicKey(pubkey)
        if len(pubkey) == 33:
            compress = True
        else:
            compress = False
        if crypted:
            return cls(
                rawkey=pubkey,
                rawvalue=None,
                pubkey=pkey.format(compressed=compress),
                sec=None,
                encryptedkey=encryptedkey,
                compressed=compress,
            )
        else:
            if len(privkey) == 279:
                sec = privkey[9:41]
            elif len(privkey) == 278:
                sec = privkey[8:40]
            else:
                sec = privkey
            prkey = PrivateKey(sec)
            if pkey == prkey.public_key:
                pkey = prkey.public_key.format(compressed=compress)
                return cls(
                    rawkey=pubkey,
                    rawvalue=privkey,
                    pubkey=pkey,
                    privkey=prkey,
                    sec=sec,
                    compressed=compress,
                )
            else:
                print("Wrong decryption password")
                return cls(
                    rawkey=pubkey,
                    rawvalue=privkey,
                    pubkey=pubkey,
                    sec=sec,
                    encryptedkey=encryptedkey,
                    compressed=compress,
                    privkey=prkey
                )

    def set_keymeta(
        self,
        version: int,
        createtime: int,
        hdkeypath: Optional[str],
        hdmasterkey: Optional[str],
        fingerprint: int,
        has_keyorigin: bool = False,
    ) -> None:
        """Set keymeta field

        :param version: version parameter
        :type version: int
        :param createtime:created time
        :type createtime: int
        :param hdkeypath: hd key path
        :type hdkeypath: str
        :param hdmasterkey: hd master key
        :type hdmasterkey: str
        :param fingerprint: fingerprint value from wallet
        :type fingerprint: int
        :param has_keyorigin: whether has keyorigin field
        :type has_keyorigin: bool
        :return: None
        :rtype:
        """
        self.version = version
        self.createtime = createtime
        self.hdkeypath = hdkeypath
        self.hdmasterkey = hdmasterkey
        self.fingerprint = fingerprint
        self.has_keyorigin = has_keyorigin

    def pubkey_towif(self, network_version: int = 0) -> bytes:
        """

        :param network_version: version byte
        :type network_version: int
        :return:
        :rtype:
        """
        prefix = bytes([network_version])
        return base58.b58encode_check(prefix + ripemd160_sha256(self.publickey))

    def privkey_towif(self, network_version: int = 0, compressed: bool = True) -> bytes:
        """

        :param network_version: version byte
        :type network_version: int
        :param compressed: whether the key is compressed
        :type compressed: bool
        :return:
        :rtype:
        """
        if self.privkey is not None:
            prefix = bytes([network_version + 128])
            if compressed:
                suffix = b"\x01"
            else:
                suffix = b""
            return base58.b58encode_check(prefix + self.privkey.secret + suffix)
        elif self.encryptedkey is not None:
            return self.encryptedkey

    def __repl__(self):
        if self.privkey is None and self.encryptedkey is not None:
            return "Pubkey: {} Encrypted Privkey: {}".format(
                self.publickey.hex(), self.encryptedkey.hex()
            )
        elif self.privkey is not None and self.encryptedkey is None:
            return "Pubkey: {} Privkey: {}".format(
                self.publickey.hex(), self.privkey.hex()
            )
        else:
            return "Pubkey: {}".format(self.publickey.hex())


def invert_txid(txid: bytes) -> str:
    """invert txid string from bytes

    :param txid: txid byte string from wallet
    :type txid: bytes
    :return: inverted txid string
    :rtype: str
    """
    tx = txid.hex()
    if len(tx) != 64:
        raise ValueError("txid %r length != 64" % tx)
    new_txid = ""
    for i in range(32):
        new_txid += tx[62 - 2 * i]
        new_txid += tx[62 - 2 * i + 1]
    return new_txid


class Transaction(object):
    """Transaction object - not to be called directly."""

    def __init__(
        self, txid: str, version: int, txin: List, txout: List, locktime: int, tx: bytes
    ) -> None:
        """

        :param txid: transaction id string
        :type txid: str
        :param version: version byte
        :type version: int
        :param txin: txin list of dicts
        :type txin: list
        :param txout: txout list of dicts
        :type txout: list
        :param locktime: int
        :type locktime:
        :param tx: bytes
        :type tx:
        """
        self.txid = txid
        self.version = version
        self.txin = txin
        self.txout = txout
        self.locktime = locktime
        self.tx = tx

    @classmethod
    def parse(cls, txid, vds):
        start = vds.read_cursor
        version = vds.read_int32()
        n_vin = vds.read_compact_size()
        txin = []
        for _ in range(n_vin):
            d = {
                "prevout_hash": vds.read_bytes(32).hex(),
                "prevout_n": vds.read_uint32(),
                "scriptSig": vds.read_bytes(vds.read_compact_size()).hex(),
                "sequence": vds.read_uint32(),
            }
            txin.append(d)
        n_vout = vds.read_compact_size()
        txout = []
        for _ in range(n_vout):
            d = {
                "value": vds.read_int64() / 1e8,
                "scriptPubKey": vds.read_bytes(vds.read_compact_size()).hex(),
            }
            txout.append(d)
        locktime = vds.read_uint32()
        tx = vds.input[start : vds.read_cursor]
        return cls(
            txid,
            version,
            sorted(txin, key=lambda i: (i["prevout_n"], i["sequence"])),
            txout,
            locktime,
            tx,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""CLI Interface to Walletlib

This is a simple implementation to allow walletlib to be used from the cli.
It will certainly gain more features as they are added. Currently it takes a wallet.dat file and dumps either a full
seet of its contents or just the keys out.

"""
import click
from walletlib import Walletdat, ProtobufWallet
import json

@click.command()
@click.argument("filename", type=click.Path(exists=True))
@click.option("-p", "--password", help="Password if any", type=click.STRING)
@click.option(
    "-o", "--output", help="File to save to. If not set, results only will be displayed"
)
@click.option(
    "-v",
    "--versionprefix",
    type=int,
    help="Force output to use this p2pkh version byte",
)
@click.option(
    "-s", "--secretprefix", type=int, help="Force output to use this WIF version byte"
)
@click.option("--keys", is_flag=True, help="Only dump keys.")
def main(filename, password, output, versionprefix, secretprefix, keys):
    if filename.endswith(".dat"):
        w = Walletdat.load(filename)
        click.echo("Loaded file")
        if password:
            w.parse(passphrase=str(password))
        else:
            w.parse()
        click.echo(
            "Found {} keypairs and {} transactions".format(len(w.keypairs), len(w.txes))
        )
        click.echo("Default version byte: {}".format(w.default_wifnetwork))
        if keys:
            if not output:
                d = w.dump_keys(version=versionprefix, privkey_prefix_override=secretprefix)
                click.echo(json.dumps(d, sort_keys=True, indent=4))
            else:
                w.dump_keys(output, version=versionprefix, privkey_prefix_override=secretprefix)
        else:
            if not output:
                d = w.dump_all(version=versionprefix, privkey_prefix_override=secretprefix)
                click.echo(json.dumps(d, sort_keys=True, indent=4))
            else:
                w.dump_all(output, version=versionprefix, privkey_prefix_override=secretprefix)
        click.echo("Done")
    else:
        try:
            w = ProtobufWallet.load(filename)
            click.echo("Loaded file")
            if password:
                w.parse(passphrase=str(password))
            else:
                w.parse()
            click.echo("Found {} keypairs and {} transactions".format(len(w.keypairs), len(w.txes)))
            click.echo("Default version byte: {}".format(w.default_wifnetwork))
            if keys:
                if not output:
                    d = w.dump_keys()
                    click.echo(json.dumps(d, sort_keys=True, indent=4))
                else:
                    w.dump_keys(output)
            else:
                if not output:
                    d = w.dump_all()
                    click.echo(json.dumps(d, sort_keys=True, indent=4))
                else:
                    w.dump_all(output)
            click.echo("Done")
        except:
            click.echo("Error, cannot read wallet file")




"""
Library for accessing cryptocurrency wallet files

keywords: wallet, crypto, wallet.dat
author: jim zhou jimtje@gmail.com
"""

from .walletdat import Walletdat
from .protobufwallet import ProtobufWallet
__version__ = "0.2.3"

__url__ = "https://github.com/jimtje/walletlib"


__all__ = ["Walletdat", "ProtobufWallet"]



import struct
import socket
from .exceptions import *


class BCDataStream(object):
    """BCDataStream from pywallet"""

    def __init__(self, input):
        self.input = bytes(input)
        self.read_cursor = 0

    def read_string(self):
        # Strings are encoded depending on length:
        # 0 to 252 :	1-byte-length followed by bytes (if any)
        # 253 to 65,535 : byte'253' 2-byte-length followed by bytes
        # 65,536 to 4,294,967,295 : byte '254' 4-byte-length followed by bytes
        # ... and the Bitcoin client is coded to understand:
        # greater than 4,294,967,295 : byte '255' 8-byte-length followed by bytes of string
        # ... but I don't think it actually handles any strings that big.
        try:
            length = self.read_compact_size()
        except IndexError:
            raise SerializationError("attempt to read past end of buffer")

        return self.read_bytes(length)

    def read_bytes(self, length):
        try:
            result = self.input[self.read_cursor : self.read_cursor + length]
            self.read_cursor += length
            return result
        except IndexError:
            raise SerializationError("attempt to read past end of buffer")

    def read_boolean(self):
        return self.read_bytes(1)[0] != chr(0)

    def read_int16(self):
        return self._read_num("<h")

    def read_uint16(self):
        return self._read_num("<H")

    def read_int32(self):
        return self._read_num("<i")

    def read_uint32(self):
        return self._read_num("<I")

    def read_int64(self):
        return self._read_num("<q")

    def read_uint64(self):
        return self._read_num("<Q")

    def read_uint256(self):
        r = 0
        for i in range(8):
            t = struct.unpack("<I", self.read_bytes(4))[0]
            r += t << (i * 32)
        return r

    def read_vector(self, f):
        nit = self.read_compact_size()
        r = []
        for _ in range(nit):
            t = f()
            t.deserialize(self)
            r.append(t)
        return r

    def read_compact_size(self):
        size = int(self.input[self.read_cursor])
        self.read_cursor += 1
        if size == 253:
            size = self._read_num("<H")
        elif size == 254:
            size = self._read_num("<I")
        elif size == 255:
            size = self._read_num("<Q")
        return size

    def _read_num(self, format):
        (i,) = struct.unpack_from(format, self.input, self.read_cursor)
        self.read_cursor += struct.calcsize(format)
        return i

    def get_onebyte(self):
        byte = self.read_bytes(1)
        self.read_cursor -= 1
        return byte


def parse_CAddress(vds):
    d = {"ip": "0.0.0.0", "port": 0, "nTime": 0}
    try:
        d["nVersion"] = vds.read_int32()
        d["nTime"] = vds.read_uint32()
        d["nServices"] = vds.read_uint64()
        d["pchReserved"] = vds.read_bytes(12)
        d["ip"] = socket.inet_ntoa(vds.read_bytes(4))
        d["port"] = vds.read_uint16()
    except:
        pass
    return d


def parse_BlockLocator(vds):
    d = {"hashes": []}
    nHashes = vds.read_compact_size()
    for i in range(nHashes):
        d["hashes"].append(vds.read_bytes(32))
    return d


def privkey_to_secret(privkey: bytes) -> bytes:
    if len(privkey) == 279:
        return privkey[9:41]
    else:
        return privkey[8:40]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: coinomiwallet.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List

import betterproto


class KeyType(betterproto.Enum):
    ORIGINAL = 1
    ENCRYPTED_SCRYPT_AES = 2
    DETERMINISTIC_MNEMONIC = 3
    DETERMINISTIC_KEY = 4


class TransactionConfidenceType(betterproto.Enum):
    UNKNOWN = 0
    BUILDING = 1
    PENDING = 2
    DEAD = 3


class TransactionConfidenceSource(betterproto.Enum):
    SOURCE_UNKNOWN = 0
    SOURCE_NETWORK = 1
    SOURCE_SELF = 2
    SOURCE_TRUSTED = 3


class TransactionPool(betterproto.Enum):
    UNSPENT = 4
    SPENT = 5
    DEAD = 10
    PENDING = 16


class WalletEncryptionType(betterproto.Enum):
    UNENCRYPTED = 1
    ENCRYPTED_SCRYPT_AES = 2
    ENCRYPTED_AES = 3


@dataclass
class PeerAddress(betterproto.Message):
    ip_address: bytes = betterproto.bytes_field(1)
    port: int = betterproto.uint32_field(2)
    services: int = betterproto.uint64_field(3)


@dataclass
class EncryptedData(betterproto.Message):
    initialisation_vector: bytes = betterproto.bytes_field(1)
    encrypted_private_key: bytes = betterproto.bytes_field(2)


@dataclass
class DeterministicKey(betterproto.Message):
    """
    * Data attached to a Key message that defines the data needed by the BIP32
    deterministic key hierarchy algorithm.
    """

    # Random data that allows us to extend a key. Without this, we can't figure
    # out the next key in the chain and should just treat it as a regular
    # ORIGINAL type key.
    chain_code: bytes = betterproto.bytes_field(1)
    # The path through the key tree. Each number is encoded in the standard form:
    # high bit set for private derivation and high bit unset for public
    # derivation.
    path: List[int] = betterproto.uint32_field(2)
    # How many children of this key have been issued, that is, given to the user
    # when they requested a fresh key? For the parents of keys being handed out,
    # this is always less than the true number of children: the difference is
    # called the lookahead zone. These keys are put into Bloom filters so we can
    # spot transactions made by clones of this wallet - for instance when
    # restoring from backup or if the seed was shared between devices. If this
    # field is missing it means we're not issuing subkeys of this key to users.
    issued_subkeys: int = betterproto.uint32_field(3)
    lookahead_size: int = betterproto.uint32_field(4)
    # * Flag indicating that this key is a root of a following chain. This chain
    # is following the next non-following chain. Following/followed chains
    # concept is used for married keychains, where the set of keys combined
    # together to produce a single P2SH multisignature address
    is_following: bool = betterproto.bool_field(5)


@dataclass
class Key(betterproto.Message):
    """
    * A key used to control Bitcoin spending. Either the private key, the
    public key or both may be present.  It is recommended that if the private
    key is provided that the public key is provided too because deriving it is
    slow. If only the public key is provided, the key can only be used to watch
    the blockchain and verify transactions, and not for spending.
    """

    type: "KeyType" = betterproto.enum_field(1)
    # Either the private EC key bytes (without any ASN.1 wrapping), or the
    # deterministic root seed. If the secret is encrypted, or this is a "watching
    # entry" then this is missing.
    secret_bytes: bytes = betterproto.bytes_field(2)
    # If the secret data is encrypted, then secret_bytes is missing and this
    # field is set.
    encrypted_data: "EncryptedData" = betterproto.message_field(3)
    # The public EC key derived from the private key. We allow both to be stored
    # to avoid mobile clients having to do lots of slow EC math on startup. For
    # DETERMINISTIC_MNEMONIC entries this is missing.
    public_key: bytes = betterproto.bytes_field(4)
    # User-provided label associated with the key.
    label: str = betterproto.string_field(5)
    deterministic_key: "DeterministicKey" = betterproto.message_field(6)


@dataclass
class TransactionInput(betterproto.Message):
    # Hash of the transaction this input is using.
    transaction_out_point_hash: bytes = betterproto.bytes_field(1)
    # Index of transaction output used by this input.
    transaction_out_point_index: int = betterproto.uint32_field(2)
    # Script that contains the signatures/pubkeys.
    script_bytes: bytes = betterproto.bytes_field(3)
    # Sequence number. Currently unused, but intended for contracts in future.
    sequence: int = betterproto.uint32_field(4)
    # Value of connected output, if known
    value: int = betterproto.int64_field(5)


@dataclass
class TransactionOutput(betterproto.Message):
    value: int = betterproto.int64_field(1)
    script_bytes: bytes = betterproto.bytes_field(2)
    # If spent, the hash of the transaction doing the spend.
    spent_by_transaction_hash: bytes = betterproto.bytes_field(3)
    # If spent, the index of the transaction input of the transaction doing the
    # spend.
    spent_by_transaction_index: int = betterproto.int32_field(4)


@dataclass
class TransactionConfidence(betterproto.Message):
    """
    * A description of the confidence we have that a transaction cannot be
    reversed in the future. Parsing should be lenient, since this could change
    for different applications yet we should maintain backward compatibility.
    """

    # This is optional in case we add confidence types to prevent parse errors -
    # backwards compatible.
    type: "TransactionConfidenceType" = betterproto.enum_field(1)
    # If type == BUILDING then this is the chain height at which the transaction
    # was included.
    appeared_at_height: int = betterproto.int32_field(2)
    # If set, hash of the transaction that double spent this one into oblivion. A
    # transaction can be double spent by multiple transactions in the case of
    # several inputs being re-spent by several transactions but we don't bother
    # to track them all, just the first. This only makes sense if type = DEAD.
    overriding_transaction: bytes = betterproto.bytes_field(3)
    # If type == BUILDING then this is the depth of the transaction in the
    # blockchain. Zero confirmations: depth = 0, one confirmation: depth = 1 etc.
    depth: int = betterproto.int32_field(4)
    broadcast_by: List["PeerAddress"] = betterproto.message_field(5)
    source: "TransactionConfidenceSource" = betterproto.enum_field(6)


@dataclass
class Transaction(betterproto.Message):
    # See Wallet.java for detailed description of pool semantics
    version: int = betterproto.int32_field(1)
    time: int = betterproto.int32_field(11)
    hash: bytes = betterproto.bytes_field(2)
    # If pool is not present, that means either:  - This Transaction is either
    # not in a wallet at all (the proto is re-used elsewhere)  - Or it is stored
    # but for other purposes, for example, because it is the overriding
    # transaction of a double spend.  - Or the Pool enum got a new value which
    # your software is too old to parse.
    pool: "TransactionPool" = betterproto.enum_field(3)
    lock_time: int = betterproto.uint32_field(4)
    updated_at: int = betterproto.int64_field(5)
    transaction_input: List["TransactionInput"] = betterproto.message_field(6)
    transaction_output: List["TransactionOutput"] = betterproto.message_field(7)
    # A list of blocks in which the transaction has been observed (on any chain).
    # Also, a number used to disambiguate ordering within a block.
    block_hash: List[bytes] = betterproto.bytes_field(8)
    block_relativity_offsets: List[int] = betterproto.int32_field(9)
    # Data describing where the transaction is in the chain.
    confidence: "TransactionConfidence" = betterproto.message_field(10)
    token_id: int = betterproto.int32_field(12)


@dataclass
class AddressStatus(betterproto.Message):
    address: str = betterproto.string_field(1)
    status: str = betterproto.string_field(2)


@dataclass
class WalletPocket(betterproto.Message):
    """* A wallet pocket"""

    network_identifier: str = betterproto.string_field(1)
    # A UTF8 encoded text description of the wallet that is intended for end user
    # provided text.
    description: str = betterproto.string_field(2)
    key: List["Key"] = betterproto.message_field(3)
    # The SHA256 hash of the head of the best chain seen by this wallet.
    last_seen_block_hash: bytes = betterproto.bytes_field(4)
    # The height in the chain of the last seen block.
    last_seen_block_height: int = betterproto.uint32_field(5)
    last_seen_block_time_secs: int = betterproto.int64_field(6)
    transaction: List["Transaction"] = betterproto.message_field(7)
    address_status: List["AddressStatus"] = betterproto.message_field(8)


@dataclass
class ScryptParameters(betterproto.Message):
    """
    * The parameters used in the scrypt key derivation function.  The default
    values are taken from http://www.tarsnap.com/scrypt/scrypt-slides.pdf.
    They can be increased - n is the number of iterations performed and  r and
    p can be used to tweak the algorithm - see:
    http://stackoverflow.com/questions/11126315/what-are-optimal-scrypt-work-
    factors
    """

    salt: bytes = betterproto.bytes_field(1)
    n: int = betterproto.int64_field(2)
    r: int = betterproto.int32_field(3)
    p: int = betterproto.int32_field(4)


@dataclass
class Wallet(betterproto.Message):
    """* A bitcoin wallet"""

    # The version number of the wallet - used to detect wallets that were
    # produced in the future (i.e the wallet may contain some future format this
    # protobuf/ code does not know about)
    version: int = betterproto.int32_field(1)
    seed: "Key" = betterproto.message_field(2)
    master_key: "Key" = betterproto.message_field(3)
    encryption_type: "WalletEncryptionType" = betterproto.enum_field(4)
    encryption_parameters: "ScryptParameters" = betterproto.message_field(5)
    pockets: List["WalletPocket"] = betterproto.message_field(6)



import hashlib
from Crypto.Cipher import AES


def doublesha256(bytestring):
    return hashlib.sha256(hashlib.sha256(bytestring).digest()).digest()


def doublesha256_checksum(bytestring):
    return doublesha256(bytestring)[:4]


def ripemd160_sha256(key):
    return hashlib.new("ripemd160", hashlib.sha256(key).digest()).digest()


class Crypter(object):

    def __init__(self):
        self.chKey = None
        self.chIV = None

    @staticmethod
    def append_PKCS7_padding(s):
        """return s padded to a multiple of 16-bytes by PKCS7 padding"""
        numpads = 16 - (len(s) % 16)
        return s + numpads * chr(numpads)

    def keyfrompassphrase(
        self, vKeyData, vSalt, nDerivIterations, nDerivationMethod
    ):
        if nDerivationMethod != 0:
            return 0
        data = vKeyData + vSalt
        for i in range(nDerivIterations):
            data = hashlib.sha512(data).digest()
        self.SetKey(data[0:32])
        self.SetIV(data[32: 32 + 16])
        return len(data)

    def SetKey(self, key):
        self.chKey = key

    def SetIV(self, iv):
        self.chIV = iv[0:16]

    def encrypt(self, data):
        return AES.new(self.chKey, AES.MODE_CBC, self.chIV).encrypt(
            Crypter.append_PKCS7_padding(data)
        )

    def decrypt(self, data):
        return AES.new(self.chKey, AES.MODE_CBC, self.chIV).decrypt(data)[0:32]





# Copyright (c) 2017 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


BECH32_VERSION_SET = ('bc', 'tb', 'bcrt')
CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"


def bech32_polymod(values):
    """Internal function that computes the Bech32 checksum."""
    generator = [0x3b6a57b2, 0x26508e6d, 0x1ea119fa, 0x3d4233dd, 0x2a1462b3]
    chk = 1
    for value in values:
        top = chk >> 25
        chk = (chk & 0x1ffffff) << 5 ^ value
        for i in range(5):
            chk ^= generator[i] if ((top >> i) & 1) else 0
    return chk


def bech32_hrp_expand(hrp):
    """Expand the HRP into values for checksum computation."""
    return [ord(x) >> 5 for x in hrp] + [0] + [ord(x) & 31 for x in hrp]


def bech32_verify_checksum(hrp, data):
    """Verify a checksum given HRP and converted data characters."""
    return bech32_polymod(bech32_hrp_expand(hrp) + data) == 1


def bech32_create_checksum(hrp, data):
    """Compute the checksum values given HRP and data."""
    values = bech32_hrp_expand(hrp) + data
    polymod = bech32_polymod(values + [0, 0, 0, 0, 0, 0]) ^ 1
    return [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]


def bech32_encode(hrp, data):
    """Compute a Bech32 string given HRP and data values."""
    combined = data + bech32_create_checksum(hrp, data)
    return hrp + '1' + ''.join([CHARSET[d] for d in combined])


def bech32_decode(bech):
    """Validate a Bech32 string, and determine HRP and data."""
    if ((any(ord(x) < 33 or ord(x) > 126 for x in bech)) or
            (bech.lower() != bech and bech.upper() != bech)):
        return (None, None)
    bech = bech.lower()
    pos = bech.rfind('1')
    if pos < 1 or pos + 7 > len(bech) or len(bech) > 90:
        return (None, None)
    if not all(x in CHARSET for x in bech[pos+1:]):
        return (None, None)
    hrp = bech[:pos]
    data = [CHARSET.find(x) for x in bech[pos+1:]]
    if not bech32_verify_checksum(hrp, data):
        return (None, None)
    return (hrp, data[:-6])


def convertbits(data, frombits, tobits, pad=True):
    """General power-of-2 base conversion."""
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            return None
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        return None
    return ret


# def decode(hrp, addr):
def decode(addr):
    """Decode a segwit address."""
    hrpgot, data = bech32_decode(addr)
    # if hrpgot != hrp:
    if hrpgot not in BECH32_VERSION_SET:
        return (None, None)
    decoded = convertbits(data[1:], 5, 8, False)
    if decoded is None or len(decoded) < 2 or len(decoded) > 40:
        return (None, None)
    if data[0] > 16:
        return (None, None)
    if data[0] == 0 and len(decoded) != 20 and len(decoded) != 32:
        return (None, None)
    return (data[0], decoded)


def encode(hrp, witver, witprog):
    """Encode a segwit address."""
    ret = bech32_encode(hrp, [witver] + convertbits(witprog, 8, 5))
    if decode(ret) == (None, None):
        return None
    return ret



class WalletDatError(Exception):

    def __init__(self, file=None, message=None):
        if message is None:
            message = "Error processing {}".format(file)
        super(WalletDatError, self).__init__(message)
        self.file = file

class SerializationError(WalletDatError):
    pass

class DatabaseError(WalletDatError):
    pass

class KeypairError(WalletDatError):
    pass

class PasswordError(WalletDatError):
    pass


from setuptools import setup, find_packages
import os

with open("requirements.txt") as f:
    requirements = f.read().splitlines()


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="walletlib",
    version="0.2.3",
    packages=find_packages(),
    include_package_data=True,
    url="https://github.com/jimtje/walletlib",
    license="Unlicense",
    author="jim zhou",
    author_email="jimtje@gmail.com",
    description="Library for accessing cryptocurrency wallet files",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='wallet crypto cryptocurrency',
    python_requires=">=3.7.0",
    install_requires=requirements,
    entry_points={
        "console_scripts": ["dumpwallet = walletlib.scripts.dumpwallet:main"]
    },
)
